from typing import Union

from ohgo.exceptions import OHGOException
from ohgo.rest_adapter import RestAdapter
from PIL import Image


class ImageHandler:
    """
    ImageHandler is a class for handling image fetching from URLs

    Attributes:
    _rest_adapter: RestAdapter for making HTTP requests to the OHGO API

    Methods:
    fetch: Fetches an image from a URL
    """

    def __init__(self, rest_adapter: RestAdapter):
        """
        Constructor for ImageHandler. Initializes the RestAdapter for making HTTP requests to the OHGO API.
        :param rest_adapter: RestAdapter for making HTTP requests to the OHGO API
        """
        self._rest_adapter = rest_adapter

    def fetch(self, url: str) -> Union[Image.Image, None]:
        """
        Fetches an image from a URL. Returns None if the image cannot be fetched.
        :param url: A string URL to fetch the image from.
        :return: A PIL Image object
        """
        try:
            image_bytes = self._rest_adapter.get_image(url)
        except OHGOException:
            return None
        return Image.open(image_bytes)



from typing import Any, List

from ohgo.models.base_model import BaseModel
from ohgo.models.models import from_list, from_str, from_float, from_int, Link, to_float, to_class


class DangerousSlowdown(BaseModel):
    """
    DangerousSlowdown is a class for storing dangerous slowdown information.

    Attributes:
    normal_mph: The normal speed limit in MPH.
    current_mph: The current speed limit in MPH.
    route_name: The name of the route where the slowdown is located.
    direction: The direction the slowdown is affecting.
    """

    normal_mph: float
    current_mph: float
    route_name: str
    direction: str

    def __init__(self, links: List['Link'], id: str, latitude: float, longitude: float, location: str, description: str, normal_mph: float, current_mph: float, route_name: str, direction: str) -> None:
        super().__init__(links, id, latitude, longitude, location, description)  # Call the parent class initializer
        self.normal_mph = normal_mph
        self.current_mph = current_mph
        self.route_name = route_name
        self.direction = direction

    @staticmethod
    def from_dict(obj: Any) -> 'DangerousSlowdown':
        base_model = BaseModel.from_base_dict(obj)  # Reuse the parent class method for base fields
        normal_mph = from_float(obj.get("normalMPH"))
        current_mph = from_float(obj.get("currentMPH"))
        route_name = from_str(obj.get("routeName"))
        direction = from_str(obj.get("direction"))
        return DangerousSlowdown(base_model.links, base_model.id, base_model.latitude, base_model.longitude, base_model.location, base_model.description, normal_mph, current_mph, route_name, direction)

    def to_dict(self) -> dict:
        result = self.base_to_dict()  # Get the base attributes
        result.update({
            "normalMPH": from_float(self.normal_mph),
            "currentMPH": from_float(self.current_mph),
            "routeName": from_str(self.route_name),
            "direction": from_str(self.direction)
        })
        return result



from dataclasses import dataclass
from typing import List, Any

from ohgo.models.models import from_list, from_str, from_float, Link, to_class, to_float

@dataclass
class BaseModel:

    links: List[Link]
    id: str
    latitude: float
    longitude: float
    location: str
    description: str

    def __init__(self, links: List['Link'], id: str, latitude: float, longitude: float, location: str, description: str) -> None:
        self.links = links
        self.id = id
        self.latitude = latitude
        self.longitude = longitude
        self.location = location
        self.description = description

    @staticmethod
    def from_base_dict(obj: Any) -> 'BaseModel':
        assert isinstance(obj, dict)
        links = from_list(Link.from_dict, obj.get("links"))
        id = from_str(obj.get("id"))
        latitude = from_float(obj.get("latitude"))
        longitude = from_float(obj.get("longitude"))
        location = from_str(obj.get("location"))
        description = from_str(obj.get("description"))
        return BaseModel(links, id, latitude, longitude, location, description)

    def base_to_dict(self) -> dict:
        result: dict = {
            "links": from_list(lambda x: to_class(Link, x), self.links),
            "id": from_str(self.id),
            "latitude": to_float(self.latitude),
            "longitude": to_float(self.longitude),
            "location": from_str(self.location),
            "description": from_str(self.description)
        }
        return result



from dataclasses import dataclass
from typing import Any, List

from ohgo.models.base_model import BaseModel
from ohgo.models.models import Link, from_list, from_str, from_float, to_float, to_class, from_int


@dataclass
class TravelDelay(BaseModel):
    """
    TravelDelay is a class for storing travel delay information.

    Attributes:
    direction: The direction of the travel delay.
    route_name: The name of the route where the travel delay is located.
    travel_time: The travel time of the route.
    delay_time: The delay time of the route.
    start_mile_marker: The starting mile marker of the route.
    end_mile_marker: The ending mile marker of the route.
    current_avg_speed: The current average speed of the route.
    normal_avg_speed: The normal average speed of the route.
    """

    direction: str
    route_name: str
    travel_time: float
    delay_time: float
    start_mile_marker: float
    end_mile_marker: float
    current_avg_speed: float
    normal_avg_speed: float

    def __init__(self, links: List[Link], id: str, latitude: float, longitude: float, location: str, description: str, direction: str, route_name: str, travel_time: float, delay_time: float, start_mile_marker: float, end_mile_marker: float, current_avg_speed: float, normal_avg_speed: float) -> None:
        super().__init__(links, id, latitude, longitude, location, description)  # Call the parent class initializer
        self.direction = direction
        self.route_name = route_name
        self.travel_time = travel_time
        self.delay_time = delay_time
        self.start_mile_marker = start_mile_marker
        self.end_mile_marker = end_mile_marker
        self.current_avg_speed = current_avg_speed
        self.normal_avg_speed = normal_avg_speed

    @staticmethod
    def from_dict(obj: Any) -> 'TravelDelay':
        base_model = BaseModel.from_base_dict(obj)  # Reuse BaseModel for shared fields
        direction = from_str(obj.get("direction"))
        route_name = from_str(obj.get("routeName"))
        travel_time = from_float(obj.get("travelTime"))
        delay_time = from_float(obj.get("delayTime"))
        start_mile_marker = from_float(obj.get("startMileMarker"))
        end_mile_marker = from_float(obj.get("endMileMarker"))
        current_avg_speed = from_float(obj.get("currentAvgSpeed"))
        normal_avg_speed = from_float(obj.get("normalAvgSpeed"))
        return TravelDelay(base_model.links, base_model.id, base_model.latitude, base_model.longitude, base_model.location, base_model.description, direction, route_name, travel_time, delay_time, start_mile_marker, end_mile_marker, current_avg_speed, normal_avg_speed)

    def to_dict(self) -> dict:
        result = self.base_to_dict()  # Get the base attributes
        result.update({
            "direction": from_str(self.direction),
            "routeName": from_str(self.route_name),
            "travelTime": to_float(self.travel_time),
            "delayTime": to_float(self.delay_time),
            "startMileMarker": to_float(self.start_mile_marker),
            "endMileMarker": to_float(self.end_mile_marker),
            "currentAvgSpeed": from_float(self.current_avg_speed),
            "normalAvgSpeed": from_float(self.normal_avg_speed)
        })
        return result


from enum import Enum


class Region(Enum):
    """
    Region is an Enum for the regions of Ohio that can be queried from the OHGO API.

    Attributes:
    AKRON: Akron region
    CINCINNATI: Cincinnati region
    CLEVELAND: Cleveland region
    COLUMBUS: Columbus region
    DAYTON: Dayton region
    TOLEDO: Toledo region
    CENTRAL_OHIO: Central Ohio region
    NE_OHIO: Northeast Ohio region
    NW_OHIO: Northwest Ohio region
    SE_OHIO: Southeast Ohio region
    SW_OHIO: Southwest Ohio region

    """
    AKRON = "akron"
    CINCINNATI = "cincinnati"
    CLEVELAND = "cleveland"
    COLUMBUS = "columbus"
    DAYTON = "dayton"
    TOLEDO = "toledo"
    CENTRAL_OHIO = "central-ohio"
    NE_OHIO = "ne-ohio"
    NW_OHIO = "nw-ohio"
    SE_OHIO = "se-ohio"
    SW_OHIO = "sw-ohio"


class SignType(Enum):
    """
    SignType is an Enum for the types of digital signs that can be queried from the OHGO API.

    Attributes:
    DMS: Dynamic Message Sign
    MESSAGE_BOARD: Message Board
    DDMS: Dynamic Digital Message Sign
    TRAVEL_TIME: Travel Time
    SIGN_QUEUE: Sign Queue
    SLOW_TRAFFIC: Slow Traffic
    VSL: Variable Speed Limit

    """
    DMS = "dms"
    MESSAGE_BOARD = "message-board"
    DDMS = "ddms"
    TRAVEL_TIME = "travel-time"
    SIGN_QUEUE = "sign-queue"
    SLOW_TRAFFIC = "slow-traffic"
    VSL = "vsl"



from datetime import datetime
from typing import *

import dateutil.parser

T = TypeVar("T")


class Result:
    """
    Result is a class for storing the results of an OHGO API query.

    Attributes:
    status_code: The status code of the query as an integer
    message: The message returned from the query
    links: The links returned from the query
    total_result_count: The total number of results returned from the query
    data: The results returned from the query. Each result is a dictionary. Default is an empty list
    rejected_filters: The rejected filters returned from the query. Default is an empty list
    _next_page: The next page of results

    Methods:
    next_page: Returns the next page of results
    """

    _next_page: str = None

    def __init__(self, status_code: int, message: str = "", data: Dict = None):
        """
        Initializes the Result object with the status code, message, and data.
        :param status_code: The response status code
        :param message: The response message
        :param data: The response data
        """
        self.status_code = int(status_code)
        self.message = str(message)
        self.links = data['links']
        self.total_result_count = data['totalResultCount']
        self.data = data['results'] if data else []
        self.rejected_filters = data['rejectedFilters']

    @property
    def next_page(self):
        """
        Returns the next page of results
        :return: If there is a next page, returns the URL of the next page. Otherwise, returns None
        """
        if not self._next_page:
            for link in self.links:
                if link['rel'] == 'next-page':
                    self._next_page = link['href']
        return self._next_page


def from_str(x: Any) -> str:
    if x is None:
        return ""
    assert isinstance(x, str)
    return x


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_float(x: Any) -> float:
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def to_float(x: Any) -> float:
    assert isinstance(x, (int, float))
    return x


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_datetime(x: Any) -> datetime:
    return dateutil.parser.parse(x)


class Link:
    """
    Link is a class for storing a link object.

    Attributes:
    href: The URL of the link
    rel: The relationship of the link to the object
    """
    href: str
    rel: str

    def __init__(self, href="", rel=""):
        """
        Initializes the Link object with the href and rel.
        :param href: The URL of the link
        :param rel: The relationship of the link to the object
        """
        self.href = href
        self.rel = rel

    @staticmethod
    def from_dict(obj: Any) -> "Link":
        assert isinstance(obj, dict)
        href = from_str(obj.get("href"))
        rel = from_str(obj.get("rel"))
        return Link(href, rel)

    def to_dict(self) -> dict:
        result: dict = {"href": from_str(self.href), "rel": from_str(self.rel)}
        return result



from dataclasses import dataclass
from typing import Any, List

from ohgo.models.base_model import BaseModel
from ohgo.models.models import Link, from_list, from_str, from_float, to_float, to_class

@dataclass
class Incident(BaseModel):
    """
    Incident is a class for storing incident information.

    Attributes:
    category: The category of the incident.
    direction: The direction related to the incident.
    route_name: The name of the route where the incident is located.
    road_status: The current status of the road.
    """

    category: str
    direction: str
    route_name: str
    road_status: str

    def __init__(self, links: List['Link'], id: str, latitude: float, longitude: float, location: str, description: str, category: str, direction: str, route_name: str, road_status: str) -> None:
        super().__init__(links, id, latitude, longitude, location, description)  # Call the parent class initializer
        self.category = category
        self.direction = direction
        self.route_name = route_name
        self.road_status = road_status

    @staticmethod
    def from_dict(obj: Any) -> 'Incident':
        base_model = BaseModel.from_base_dict(obj)  # Reuse BaseModel for shared fields
        category = from_str(obj.get("category"))
        direction = from_str(obj.get("direction"))
        route_name = from_str(obj.get("routeName"))
        road_status = from_str(obj.get("roadStatus"))
        return Incident(base_model.links, base_model.id, base_model.latitude, base_model.longitude, base_model.location, base_model.description, category, direction, route_name, road_status)

    def to_dict(self) -> dict:
        result = self.base_to_dict()  # Get the base attributes
        result.update({
            "category": from_str(self.category),
            "direction": from_str(self.direction),
            "routeName": from_str(self.route_name),
            "roadStatus": from_str(self.road_status)
        })
        return result



import logging

from PIL.Image import Image

from ohgo.models.camera import Camera, CameraView
from ohgo.models.contruction import Construction
from ohgo.models.digital_sign import DigitalSign
from ohgo.models.incident import Incident
from ohgo.models.travel_delay import TravelDelay
from ohgo.models.query_params import QueryParams, DigitalSignParams, ConstructionParams, WeatherSensorSiteParams
from ohgo.models.weather_sensor_site import WeatherSensorSite
from ohgo.models.dangerous_slowdown import DangerousSlowdown
from ohgo.rest_adapter import RestAdapter
from ohgo.exceptions import OHGOException
from ohgo.image_handler import ImageHandler
from typing import List
from functools import singledispatchmethod

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)


class OHGOClient:
    """
    OHGOClient provides methods for fetching data from OHGO including Cameras, Construction, Digital Signage,
    Incidents, Travel Delays, and Weather Sensors.

    Attributes:
    _rest_adapter: RestAdapter for making HTTP requests to the OHGO API
    _image_handler: ImageHandler for fetching images from OHGO API

    Methods:
    get_cameras: Fetches cameras from OHGO API
    get_camera: Fetches a single camera from OHGO API
    get_image: Fetches an image from a Camera or CameraView
    get_images: Fetches images from all CameraViews of a Camera
    get_digital_signs: Fetches digital signs from OHGO API
    get_digital_sign: Fetches a single digital sign from OHGO API
    get_constructions: Fetches construction from OHGO API
    get_construction: Fetches a single construction from OHGO API
    get_weather_sensor_sites: Fetches weather sensor sites from OHGO API
    get_weather_sensor_site: Fetches a single weather sensor site from OHGO API
    get_incidents: Fetches incidents from OHGO API
    get_incident: Fetches a single incident from OHGO API
    get_dangerous_slowdowns: Fetches dangerous slowdowns from OHGO API
    get_dangerous_slowdown: Fetches a single dangerous slowdown from OHGO API

    """

    def __init__(
            self,
            api_key: str,
            hostname: str = "publicapi.ohgo.com",
            ver: str = "v1",
            ssl_verify: bool = True,
            logger: logging.Logger = None,
    ):
        """
        Constructor for OHGOClient
        :param api_key: Required API key for OHGO API
        :param hostname: The hostname of the OHGO API, almost always "publicapi.ohgo.com"
        :param ver: The version of the API to use, defaults to "v1"
        :param ssl_verify: Whether to verify SSL certificates, defaults to True
        :param logger: (optional) A logger to use for logging, defaults to None
        """
        self._rest_adapter = RestAdapter(hostname, api_key, ver, ssl_verify, logger)
        self._image_handler = ImageHandler(self._rest_adapter)

    def get_cameras(self, params: QueryParams = None, fetch_all=False, **kwargs) -> List[Camera]:
        """
        Fetches cameras from the OHGO API
        :param params: QueryParams object to pass to the API
        :param fetch_all: Pages through all results if True. Recommended to use page-all param instead.
        :param kwargs: Extra arguments to pass to the API. QueryParams recommended instead. (provides basic validation)
        :return: List of Camera objects
        """
        ep_params = dict(params) if params else {}
        ep_params.update(kwargs)  # Add any extra arguments to ep_params

        result = self._rest_adapter.get(endpoint="cameras", fetch_all=fetch_all, ep_params=ep_params)
        cameras = [Camera.from_dict(camera) for camera in result.data]
        return cameras

    def get_camera(self, camera_id) -> Camera:
        """
        Fetches a single camera from the OHGO API
        :param camera_id: The ID of the camera to fetch
        :return: A Camera object
        """
        result = self._rest_adapter.get(endpoint=f"cameras/{camera_id}")
        if len(result.data) == 0:
            raise OHGOException(f"No camera found with ID {camera_id}")
        return Camera.from_dict(result.data[0])

    @singledispatchmethod
    def get_image(self, obj, size="small"):
        """
        Generic method for fetching an image from an object. Not implemented for all types.
        :param obj: The object to fetch the image from
        :param size: the size of the image to fetch, either "small" or "large"
        :return: A PIL Image object
        """
        raise NotImplementedError("Cannot get image from this type")

    @get_image.register
    def _(self, camera_view: CameraView, size="small") -> Image:
        """
        Fetches an image from a CameraView
        :param camera_view: A CameraView object.
        :param size: the size of the image to fetch, either "small" or "large"
        :return: A PIL Image object
        """
        if size == "small":
            return self._image_handler.fetch(camera_view.small_url)
        elif size == "large":
            return self._image_handler.fetch(camera_view.large_url)

    @get_image.register
    def _(self, camera: Camera, size="small") -> Image:
        """
        Fetches an image from a Camera
        :param camera: A Camera object
        :param size: the size of the image to fetch, either "small" or "large"
        :return: The first CameraView image as a PIL Image object
        """
        if len(camera.camera_views) == 0:
            raise OHGOException(f"No camera views found for camera {camera.id}")
        return self.get_image(camera.camera_views[0], size)

    @singledispatchmethod
    def get_images(self, obj) -> List[Image]:
        """
        Generic method for fetching images from an object. Not implemented for all types.
        :param obj: The object to fetch images from
        :return: List of PIL Image objects
        """
        raise NotImplementedError("Cannot get images from this type")

    @get_images.register
    def _(self, camera: Camera, size="small") -> List[Image]:
        """
        Loops through all CameraViews of a Camera and fetches images for each.
        :param camera: A Camera object
        :param size: the size of the image to fetch, either "small" or "large"
        :return: List of PIL Image objects
        """
        return [self.get_image(view, size) for view in camera.camera_views]

    @get_images.register
    def _(self, digital_sign: DigitalSign) -> List[Image]:
        """
        Fetches all images from a DigitalSign. Filters out any None values.
        :param digital_sign:
        :return: a list of PIL Image objects associated with the DigitalSign
        """
        # fetch might return None due to request exceptions, we don't want those values
        images = [self._image_handler.fetch(image_url) for image_url in digital_sign.image_urls]
        return [image for image in images if image is not None]

    def get_digital_signs(self, params: DigitalSignParams = None, fetch_all=False, **kwargs) -> List[DigitalSign]:
        """
        Fetches digital signs from the OHGO API
        :param params: DigitalSignParams object to pass to the API
        :param fetch_all: Pages through all results if True. Recommended to use page-all param instead.
        :param kwargs: Extra arguments to pass to the API.
        :return: List of DigitalSign objects
        """
        ep_params = dict(params) if params else {}
        ep_params.update(kwargs)

        result = self._rest_adapter.get(endpoint="digital-signs", fetch_all=fetch_all, ep_params=ep_params)
        digital_signs = [DigitalSign.from_dict(digital_sign) for digital_sign in result.data]
        return digital_signs

    def get_digital_sign(self, digital_sign_id) -> DigitalSign:
        """
        Fetches a single digital sign from the OHGO API
        :param digital_sign_id: The ID of the digital sign to fetch
        :return: A DigitalSign object
        """
        result = self._rest_adapter.get(endpoint=f"digital-signs/{digital_sign_id}")
        if len(result.data) == 0:
            raise OHGOException(f"No digital sign found with ID {digital_sign_id}")
        return DigitalSign.from_dict(result.data[0])

    def get_constructions(self, params: ConstructionParams = None, fetch_all=False, **kwargs) -> List[Construction]:
        """
        Fetches construction from the OHGO API
        :param params: ConstructionParams object to pass to the API
        :param fetch_all: Pages through all results if True. Recommended to use page-all param instead.
        :param kwargs: Extra arguments to pass to the API.
        :return: List of Construction objects
        """
        ep_params = dict(params) if params else {}
        ep_params.update(kwargs)

        result = self._rest_adapter.get(endpoint="construction", fetch_all=fetch_all, ep_params=ep_params)
        construction = [Construction.from_dict(construction) for construction in result.data]
        return construction

    def get_construction(self, construction_id) -> Construction:
        """
        Fetches a single construction from the OHGO API
        :param construction_id: The ID of the construction to fetch
        :return: A Construction object
        """
        result = self._rest_adapter.get(endpoint=f"construction/{construction_id}")
        if len(result.data) == 0:
            raise OHGOException(f"No construction found with ID {construction_id}")
        return Construction.from_dict(result.data[0])

    def get_weather_sensor_sites(self, params: WeatherSensorSiteParams = None, fetch_all=False, **kwargs) -> List[WeatherSensorSite]:
        """
        Fetches weather sensor sites from the OHGO API
        :param params: WeatherSensorSiteParams object to pass to the API
        :param fetch_all: Pages through all results if True. Recommended to use page-all param instead.
        :param kwargs: Extra arguments to pass to the API.
        :return: List of WeatherSensorSite objects
        """
        ep_params = dict(params) if params else {}
        ep_params.update(kwargs)

        result = self._rest_adapter.get(endpoint="weather-sensor-sites", fetch_all=fetch_all, ep_params=ep_params)
        weather_sensor_sites = [WeatherSensorSite.from_dict(weather_sensor_site) for weather_sensor_site in result.data]
        return weather_sensor_sites

    def get_weather_sensor_site(self, site_id) -> WeatherSensorSite:
        """
        Fetches a single weather sensor site from the OHGO API
        :param site_id: The ID of the weather sensor site to fetch
        :return: A WeatherSensorSite object
        """
        result = self._rest_adapter.get(endpoint=f"weather-sensor-sites/{site_id}")
        if len(result.data) == 0:
            raise OHGOException(f"No weather sensor site found with ID {site_id}")
        return WeatherSensorSite.from_dict(result.data[0])

    def get_incidents(self, params: QueryParams = None, fetch_all=False, **kwargs) -> List[Incident]:
        """
        Fetches incidents from the OHGO API
        :param params: QueryParams object to pass to the API
        :param fetch_all: Pages through all results if True. Recommended to use page-all param instead.
        :param kwargs: Extra arguments to pass to the API.
        :return: List of Incident objects
        """
        result = self._rest_adapter.get(endpoint="incidents", fetch_all=fetch_all, ep_params=dict(params) if params else kwargs)
        incidents = [Incident.from_dict(incident) for incident in result.data]
        return incidents

    def get_incident(self, incident_id) -> Incident:
        """
        Fetches a single incident from the OHGO API
        :param incident_id: The ID of the incident to fetch
        :return: An Incident object
        """
        result = self._rest_adapter.get(endpoint=f"incidents/{incident_id}")
        if len(result.data) == 0:
            raise OHGOException(f"No incident found with ID {incident_id}")
        return Incident.from_dict(result.data[0])

    def get_dangerous_slowdowns(self, params: QueryParams = None, fetch_all=False, **kwargs) -> List[DangerousSlowdown]:
        """
        Fetches dangerous slowdowns from the OHGO API
        :param params: QueryParams object to pass to the API
        :param fetch_all: Pages through all results if True. Recommended to use page-all param instead.
        :param kwargs: Extra arguments to pass to the API.
        :return: List of DangerousSlowdown objects
        """
        result = self._rest_adapter.get(endpoint="dangerous-slowdowns", fetch_all=fetch_all, ep_params=dict(params) if params else kwargs)
        slowdowns = [DangerousSlowdown.from_dict(slowdown) for slowdown in result.data]
        return slowdowns

    def get_dangerous_slowdown(self, slowdown_id) -> DangerousSlowdown:
        """
        Fetches a single dangerous slowdown from the OHGO API
        :param slowdown_id: The ID of the dangerous slowdown to fetch
        :return: A DangerousSlowdown object
        """
        result = self._rest_adapter.get(endpoint=f"dangerous-slowdowns/{slowdown_id}")
        if len(result.data) == 0:
            raise OHGOException(f"No dangerous slowdown found with ID {slowdown_id}")
        return DangerousSlowdown.from_dict(result.data[0])

    def get_travel_delays(self, params: QueryParams = None, fetch_all=False, **kwargs) -> List[TravelDelay]:
        """
        Fetches travel delays from the OHGO API
        :param params: QueryParams object to pass to the API
        :param fetch_all: Pages through all results if True. Recommended to use page-all param instead.
        :param kwargs: Extra arguments to pass to the API.
        :return: List of TravelDelay objects
        """
        result = self._rest_adapter.get(endpoint="travel-delays", fetch_all=fetch_all, ep_params=dict(params) if params else kwargs)
        delays = [TravelDelay.from_dict(delay) for delay in result.data]
        return delays

    def get_travel_delay(self, delay_id) -> TravelDelay:
        """
        Fetches a single travel delay from the OHGO API
        :param delay_id: The ID of the travel delay to fetch
        :return: A TravelDelay object
        """
        result = self._rest_adapter.get(endpoint=f"travel-delays/{delay_id}")
        if len(result.data) == 0:
            raise OHGOException(f"No travel delay found with ID {delay_id}")


from ohgo.models.models import *
from typing import Tuple
from dataclasses import dataclass

from ohgo.models.enums import *
import logging

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)


def to_enum(enum_str, enum_type) -> Union[str, Enum]:
    """
    Attempts to convert a string to an Enum. If the string is not a valid Enum, it returns the string.
    :param enum_str: The string to convert
    :param enum_type: The Enum type to convert to
    :return: An Enum or the original string
    """
    enum = enum_str
    try:
        enum = enum_type(enum_str)
    except ValueError:
        logger.warning(f"{enum_str} may not be valid {enum_type.__name__}.")
    return enum


@dataclass
class QueryParams:
    """
    QueryParams is a dataclass for storing query parameters to pass to the OHGO API.

    Attributes:
    region: The region to query, defaults to None. Format: Region enum or string
    map_bounds_sw: The southwest corner of a bounding box, defaults to None. Format: (lat, lon)
    map_bounds_ne: The northeast corner of a bounding box, defaults to None. Format: (lat, lon)
    page_size: The number of results to return per page, defaults to None. Format: int
    page: The page number to return, defaults to None. Format: int
    page_all: Whether to fetch all pages, defaults to None. Format: True or False
    radius: The radius to search around a point, defaults to None. Format: (lat, lon, radius)

    Methods:
    to_region: Converts a string to a Region enum
    to_dict: Converts the QueryParams object to a dictionary
    """

    region: Union[str, Region] = None
    map_bounds_sw: Optional[Tuple[float, float]] = None
    map_bounds_ne: Optional[tuple] = None
    page_size: Optional[int] = None
    page: Optional[int] = None
    page_all: Optional[bool] = None
    radius: Optional[Tuple[float, float, float]] = None

    def __post_init__(self) -> None:
        """
        Post-init method for QueryParams. Validates the region attribute.
        """
        if isinstance(self.region, str):
            self.region = to_enum(self.region, Region)

        if (self.map_bounds_sw and not self.map_bounds_ne) or (self.map_bounds_ne and not self.map_bounds_sw):
            logger.warning(f"Both map-bounds-sw and map-bounds-nw should be set.")

    def to_dict(self) -> dict:
        """
        Converts the QueryParams object to a dictionary
        :return: A dictionary of the QueryParams object
        """
        ep_params = {}
        for key, value in self.__dict__.items():
            if value is not None:
                if isinstance(value, Enum):
                    value = value.value
                elif isinstance(value, tuple):
                    value = ','.join(map(str, value))
                param_name = key.replace("_", "-")  # OHGO API uses '-' in query vars
                ep_params[param_name] = value
        return ep_params

    def __iter__(self):
        return iter(self.to_dict().items())

    def __getitem__(self, key):
        return self.to_dict()[key]

    def __repr__(self):
        return f"QueryParams({self.to_dict()})"


@dataclass
class DigitalSignParams(QueryParams):
    """
        DigitalSignParams is a dataclass for storing query parameters to pass to the OHGO. Has all parameters of QueryParams plus sign-type.
    """

    sign_type: Union[str, SignType] = None

    def __post_init__(self) -> None:
        """
        Post-init method for DigitalSignParams. Validates the sign_type attribute.
        """
        super().__post_init__()
        if not isinstance(self.sign_type, SignType):
            self.sign_type = to_enum(self.sign_type, SignType)

    def to_dict(self) -> dict:
        """
        Converts the DigitalSignParams object to a dictionary
        :return: A dictionary of the DigitalSignParams object
        """
        ep_params = super().to_dict()
        if self.sign_type:
            if isinstance(self.sign_type, Enum):
                sign_type = self.sign_type.value
            else:
                sign_type = self.sign_type
            ep_params["sign-type"] = sign_type
        return ep_params

    def __repr__(self):
        return f"DigitalSignParams({self.to_dict()})"

@dataclass
class ConstructionParams(QueryParams):
    """
    ConstructionParams is a dataclass for storing query parameters to pass to the OHGO API. Has all parameters of
    QueryParams plus construction-specific parameters, include_future and future_only.

    Attributes:
    include_future: Whether to include future construction, defaults to None. Format: True or False
    future_only: Whether to only include future construction, defaults to None. Format: True or False

    """

    include_future: Optional[datetime] = None
    future_only: Optional[datetime] = None

    def __post_init__(self) -> None:
        """
        Post-init method for ConstructionParams. Validates the include_future and future_only attributes.
        """
        super().__post_init__()
        if self.include_future and not isinstance(self.include_future, datetime):
            logger.warning(f"include_future should be a datetime object to automatically convert to yyyy-mm-dd")

        if self.future_only and not isinstance(self.future_only, datetime):
            logger.warning(f"future_only should be a datetime object to automatically convert to yyyy-mm-dd")
    def to_dict(self) -> dict:
        """
        Converts the ConstructionParams object to a dictionary
        :return: A dictionary of the ConstructionParams object
        """
        ep_params = super().to_dict()
        if self.include_future:
            if isinstance(self.include_future, datetime):
                include_future = self.include_future.strftime("%Y-%m-%d")
            else:
                include_future = self.include_future
            ep_params["include-future"] = include_future
        if self.future_only:
            if isinstance(self.future_only, datetime):
                future_only = self.future_only.strftime("%Y-%m-%d")
            else:
                future_only = self.future_only
            ep_params["future-only"] = future_only
        return ep_params

    def __repr__(self):
        return f"ConstructionParams({self.to_dict()})"

@dataclass
class WeatherSensorSiteParams(QueryParams):
    """
    WeatherSensorSiteParams is a dataclass for storing query parameters to pass to the OHGO API. Has all parameters
    of QueryParams plus weather sensor site-specific parameter: hazards_only.
    """

    hazards_only: Optional[bool] = None

    def to_dict(self) -> dict:
        """
        Converts the WeatherSensorSiteParams object to a dictionary
        :return: A dictionary of the WeatherSensorSiteParams object
        """
        ep_params = super().to_dict()
        if self.hazards_only:
            ep_params["hazards-only"] = self.hazards_only
        return ep_params

    def __repr__(self):
        return f"WeatherSensorSiteParams({self.to_dict()})"
